import json

from rich import box
from rich.console import Console
from rich.table import Table

from ._install import _download_and_install, _fetch_info_rest, _pick_asset
from .tools import get_dir


def update_all(token_file=None) -> int:
    token = token_file.readline().strip() if token_file else None

    fontman_dir = get_dir()

    update_list = []

    for directory in sorted(fontman_dir.glob("*")):
        if not (directory / "fontman.json").exists():
            continue

        with open(directory / "fontman.json") as f:
            d = json.load(f)

        old_tag = d["tag"]
        new_tag, assets = _fetch_info_rest(d["repo"], token=token)

        if old_tag != new_tag:
            asset = _pick_asset(assets)
            update_list.append((directory, d["repo"], old_tag, new_tag, asset))

    console = Console()

    if len(update_list) == 0:
        console.print("All installed fonts up-to-date.")
        return 0

    console.print("Available updates:")
    table = Table(show_header=False)
    table.box = box.SIMPLE
    table.add_column("Name")
    table.add_column("old")
    table.add_column("arrow")
    table.add_column("new")
    table.add_column("size")
    for _, repo, old_tag, new_tag, asset in update_list:
        size_mb = asset["size"] / 1024 / 1024
        table.add_row(f"{repo}:", old_tag, "→", new_tag, f"[{size_mb:.1f} MB]")
    console.print(table)

    console.print("Update? [Y/n] ", end="")
    choice = console.input().lower()
    if choice in ["n", "no"]:
        console.print("Abort.")
        return 1

    with console.status("Updating..."):
        for target_dir, repo, _, new_tag, asset in update_list:
            _download_and_install(target_dir, repo, asset, new_tag)
            console.print(f"Successfully updated to {repo} {new_tag}")

    return 0



import json

from rich import box
from rich.console import Console
from rich.table import Table

from .tools import get_dir


def list_fonts():
    fontman_dir = get_dir()

    table = Table(show_header=False)
    table.box = box.SIMPLE

    table.add_column("Name")
    table.add_column("Release")

    for directory in sorted(fontman_dir.glob("*")):
        if not (directory / "fontman.json").exists():
            continue

        with open(directory / "fontman.json") as f:
            d = json.load(f)

        table.add_row(d["repo"], d["tag"])

    console = Console()
    if table.row_count == 0:
        console.print(
            f"Found no fontman-installed fonts in {fontman_dir}", style="yellow"
        )
    else:
        console.print(f"Installed fonts (in {fontman_dir}):")
        console.print(table)

    return 0



from __future__ import annotations

import json
import shutil

from rich.console import Console

from .tools import get_dir, normalize_dirname


def remove(names: list[str]) -> int:
    fontman_dir = get_dir()

    scheduled_dirs = []
    for directory in fontman_dir.glob("*"):
        if not (directory / "fontman.json").exists():
            continue
        for name in names:
            if normalize_dirname(name) in directory.name.lower():
                scheduled_dirs.append(directory)
                break

    console = Console()

    if len(scheduled_dirs) == 0:
        console.print("Found no fonts to remove.", style="yellow")
        return 1

    repos = []
    for directory in sorted(scheduled_dirs):
        with open(directory / "fontman.json") as f:
            d = json.load(f)
        repos.append(d["repo"])

    console.print("The following fonts are scheduled for removal:\n")
    for repo in repos:
        console.print(f"  {repo}")

    console.print("\nRemove? \\[y/N] ", end="")
    choice = console.input().lower()
    if choice not in ["y", "yes"]:
        console.print("Abort.")
        return 1

    for repo, directory in zip(repos, scheduled_dirs):
        shutil.rmtree(directory)
        console.print(f"Successfully removed {repo}")

    return 0



import fontman


def test_install_remove(monkeypatch):
    fontman.cli(["install", "adobe-fonts/source-code-pro"])
    fontman.cli(["list"])

    # remove and abort
    # https://stackoverflow.com/a/36377194/353337
    monkeypatch.setattr("builtins.input", lambda _: "n")
    out = fontman.cli(["rm", "adobe-fonts/source-code-pro"])
    assert out == 1

    # actually remove
    monkeypatch.setattr("builtins.input", lambda _: "y")
    out = fontman.cli(["rm", "adobe-fonts/source-code-pro"])
    assert out == 0


def test_list(monkeypatch):
    # show empty
    fontman.cli(["list"])

    monkeypatch.setattr("builtins.input", lambda _: "y")
    out = fontman.cli(["rm", "adobe-fonts/source-code-pro"])
    assert out == 1


def test_update(monkeypatch):
    # nothing to update
    out = fontman.cli(["up"])
    assert out == 0

    # install an outdated version of something
    fontman.cli(["install", "ibm/plex==v5.1.3"])

    # request update and abort
    monkeypatch.setattr("builtins.input", lambda _: "n")
    out = fontman.cli(["up"])
    assert out == 1

    # actually update
    monkeypatch.setattr("builtins.input", lambda _: "y")
    out = fontman.cli(["up"])
    assert out == 0



import argparse
from importlib import metadata
from sys import version_info

from ._install import install_fonts
from ._list import list_fonts
from ._remove import remove
from ._update import update_all


def cli(argv=None):
    parser = argparse.ArgumentParser(
        description=("Manage fonts."),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=get_version_text(),
        help="display version information",
    )

    subparsers = parser.add_subparsers(title="subcommands")

    parser_install = subparsers.add_parser(
        "install", help="Install new fonts", aliases=["i", "inst", "add"]
    )
    _cli_install(parser_install)
    parser_install.set_defaults(
        func=lambda args: install_fonts(args.repos, args.token_file, args.force)
    )

    parser_list = subparsers.add_parser(
        "list", help="List installed fonts", aliases=["l"]
    )
    parser_list.set_defaults(func=lambda _: list_fonts())

    parser_remove = subparsers.add_parser(
        "remove", help="Remove fonts", aliases=["rm", "uninstall"]
    )
    _cli_remove(parser_remove)
    parser_remove.set_defaults(func=lambda args: remove(args.names))

    parser_update = subparsers.add_parser(
        "update", help="Update installed fonts", aliases=["upgrade", "up"]
    )
    _cli_update(parser_update)
    parser_update.set_defaults(func=lambda args: update_all(args.token_file))

    # args = parser.parse_args(argv)
    # token = args.token_file.readline().strip() if args.token_file else None
    # return update_all(token)

    args = parser.parse_args(argv)
    return args.func(args)


def _cli_install(parser):
    parser.add_argument(
        "repos", nargs="+", type=str, help="GitHub repository fonts to install"
    )

    parser.add_argument(
        "-t",
        "--token-file",
        type=argparse.FileType("r"),
        help="File containing a GitHub token (can be - [stdin])",
    )

    parser.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Force re-installation (default: false)",
    )

    # token = args.token_file.readline().strip() if args.token_file else None

    # for repo in args.repo:
    #     install(repo, token, args.force)


def _cli_remove(parser):
    parser.add_argument(
        "names", nargs="+", type=str, help="GitHub repository fonts to remove"
    )


def _cli_update(parser):
    parser.add_argument(
        "-t",
        "--token-file",
        type=argparse.FileType("r"),
        help="File containing a GitHub token (can be - [stdin])",
    )


def get_version_text():
    try:
        version = metadata.version("fontman")
    except Exception:
        version = "unknown"

    python_version = f"{version_info.major}.{version_info.minor}.{version_info.micro}"
    version_text = "\n".join(
        [
            f"fontman {version} [Python {python_version}]",
            "Copyright (c) 2021 Nico Schlömer <nico.schloemer@gmail.com>",
        ]
    )
    return version_text



from __future__ import annotations

import datetime
import io
import json
import shutil
import tarfile
import zipfile
from pathlib import Path

import requests
from rich.console import Console

from ._errors import FontmanError
from .tools import get_dir, normalize_dirname


def install_fonts(strings: list[str], token_file, force: bool) -> None:
    token = token_file.readline().strip() if token_file else None
    console = Console()

    with console.status("Installing...") as status:
        for string in strings:
            # separate name and version
            res = string.split("==")
            if len(res) == 1:
                name = string
                tag = None
            else:
                assert len(res) == 2
                name = res[0]
                tag = res[1]

            dirname = normalize_dirname(name)
            target_dir = get_dir() / dirname

            console = Console()

            if not force:
                db_file = target_dir / "fontman.json"
                if db_file.exists():
                    console.print(f"{name} is already installed", style="yellow")
                    return

                if target_dir.exists():
                    console.print(
                        "Target directory exists but does not contain fontman-installed font",
                        style="red",
                    )
                    return

            try:
                tag, assets = _fetch_info_rest(name, tag, token)
            except FontmanError as e:
                console.print(str(e), style="red")
                return

            asset = _pick_asset(assets)
            size_mb = asset["size"] / 1024 / 1024
            status.update(f"Installing {string} [{size_mb:.1f} MB]...")
            _download_and_install(target_dir, name, asset, tag)
            console.print(f"done ([bold]{tag}[/])")


def _pick_asset(assets):
    if len(assets) == 0:
        raise RuntimeError("Release without assets")
    elif len(assets) == 1:
        return assets[0]

    # If there are multiple assets, choose one. First, create a rating.
    ratings = [0] * len(assets)
    for k, item in enumerate(assets):
        if "otf" in item["name"].lower() or "opentype" in item["name"].lower():
            ratings[k] += 4
        elif "super-ttc" in item["name"].lower():
            # Iosevka has those super-ttc fonts
            ratings[k] += 3
        elif "ttc" in item["name"].lower():
            ratings[k] += 2
        elif "ttf" in item["name"].lower() or "truetype" in item["name"].lower():
            ratings[k] += 1

    max_rating_assets = [
        asset for asset, rating in zip(assets, ratings) if rating == max(ratings)
    ]

    # From those, chose the least specific ones, i.e., the ones with the shortest stem.
    # This satisfies the Iosevka use case where you have tons of super-ttc files
    #
    #   super-ttc-iosevka-7.3.0.zip
    #   super-ttc-iosevka-aile-7.3.0.zip
    #   super-ttc-iosevka-curly-7.3.0.zip
    #   ...
    #
    # We only want the first
    min_stem_length = min(len(Path(item["name"]).stem) for item in max_rating_assets)

    shortest_name_assets = [
        item
        for item in max_rating_assets
        if len(Path(item["name"]).stem) == min_stem_length
    ]

    # From those, pick the one with the smallest size
    return min(shortest_name_assets, key=lambda item: item["size"])


def _download_and_install(
    target_dir: Path, repo: str, asset: dict, tag_name: str
) -> None:
    url = asset["browser_download_url"]
    res = requests.get(url, stream=True)
    if not res.ok:
        raise RuntimeError(f"Failed to fetch resource from {url}")

    if target_dir.exists():
        shutil.rmtree(target_dir)

    if asset["content_type"] in ["application/zip", "application/x-zip-compressed"]:
        archive = zipfile.ZipFile(io.BytesIO(res.content), "r")
    elif asset["content_type"] == "application/x-gzip":
        archive = tarfile.open(fileobj=res.raw, mode="r|gz")
    elif asset["content_type"] == "application/x-xz":
        archive = tarfile.open(fileobj=res.raw, mode="r|xz")
    else:
        raise RuntimeError(f"Unknown content type {asset['content_type']}")

    _extract_selectively(archive, target_dir)

    # Sometimes, the archives will contain stray __MACOSX files. Remove those
    if (target_dir / "__MACOSX").exists():
        shutil.rmtree(target_dir / "__MACOSX")

    # add database file
    db = {
        "repo": repo,
        "last-updated": datetime.datetime.now().isoformat(),
        "tag": tag_name,
    }
    with open(target_dir / "fontman.json", "w") as f:
        json.dump(db, f, indent=2)


def _extract_selectively(archive, target_dir: Path) -> None:
    # If there are OTF files, only install those
    # Otherwise, if there are TTC files, only install those
    # Otherwise, if there are TTF files, only install those
    for ext in [".otf", ".ttc", "ttf"]:
        files = [item for item in archive.namelist() if item.endswith(ext)]
        if len(files) > 0:
            for item in files:
                archive.extract(item, target_dir)
            return

    # Otherwise, if there are a top-level "desktop" folders, only install those
    desktop_folders = [
        item for item in archive.namelist() if "desktop" in item.split("/")[0].lower()
    ]
    if len(desktop_folders) > 0:
        for item in archive.namelist():
            for directory in desktop_folders:
                if item.startswith(directory):
                    archive.extract(item, target_dir)
                    break
        return

    # Fallback: Just unzip the entire archive
    archive.extractall(target_dir)


def _fetch_info_rest(
    name: str, tag: str | None = None, token: str | None = None
) -> tuple[str, dict]:
    # The latest release is the most recent non-prerelease, non-draft release, sorted by
    # the created_at attribute.

    url = f"https://api.github.com/repos/{name}/releases/"
    url += "latest" if tag is None else f"tags/{tag}"

    headers = {}
    if token is not None:
        headers["Authorization"] = f"token {token}"

    res = requests.get(url, headers=headers)
    if not res.ok:
        raise FontmanError(
            "not found"
            if res.status_code == 404
            else f"failed request to {url} ({res.status_code}, {res.reason})"
        )

    res_json = res.json()

    return res_json["tag_name"], res_json["assets"]


# def _fetch_info_graphql(repo):
#     url = "https://api.github.com/graphql"
#
#     # headers = {"Authorization": "token TOKEN"}
#
#     owner, name = repo.split("/")
#
#     query = f"""
#     {{
#       repository(owner: "{owner}", name: "{name}") {{
#         latestRelease
#     }}
#     """
#     res = requests.post(url,json={"query": query})
#
#     if not res.ok:
#         raise RuntimeError(f"Failed request to {url} ({res.status_code}, {res.reason})")
#
#     print(res.json())
#
#     exit(1)
#
#     return res_json["tag_name"], res_json["assets"]



from pathlib import Path

import appdirs


def normalize_dirname(string: str) -> str:
    return string.replace("/", "_").lower()


def get_dir() -> Path:
    fontman_dir = Path(appdirs.user_data_dir()) / "fonts" / "fontman"
    fontman_dir.mkdir(parents=True, exist_ok=True)
    return fontman_dir
